#! /usr/bin/perl -w

# generate_boximage_BMC.pl: this script generates an image and and
# associated web image map, for display on a webpage (not provided).
# The input is a file containing information about mRNA expression
# intensity and protein detection information.
#
# The file is tab-delimited, and formatted like so...
#
# # B number  mRNA intensity  mRNA present  Protein detected - total list Operon ID
# 1 b0001 1344  P   1
# 2 b0002 13624 P P 1
# 3 b0003 6781  P P 1
#
# Originally, the gene descriptions that appear over the image were
# from a database, but to facilitate distribution of this code, I
# converted it to pull descriptions from a text file, which is also
# included.
#
# This software requires the Perl module 'GD' in order to work.  You
# will need to make sure this module is installed on your system.
#
# Files in archive:
# ARTISTIC                               - The license this code is distrubuted under
# KUSTU_Data-for-GI.txt                  - The data from the used in the publication
# ecoli_descriptions_protein_profile.txt - ecoli gene descriptions, from the
#                                          Ecoli Entry Point (http://coli.berkeley.edu/ecoli)
# generate_boximage_BMC.pl                - this perl script, which generates the image/imagemap
#
# This software was written by me, Brian C. Thomas, in 2004.  I was funded
# to do this project by Sydney Kustu at the University of California,
# Berkeley.
#
# N.B.  I cannot support this code and, as stated in the license, I am
# providing it as is.  It should work on any webserver with perl
# running either as Mod_Perl, or as a just plain CGI.
#




( @ARGV > 0 ) or usage();

use strict;
use GD;

my $gene_desc_file = "ecoli_descriptions_protein_profile.txt";

my %PARAMS = ();
my %COLORS = ();

# parse args
parse_args( \@ARGV, \%PARAMS);

# start file processing
my $rc = process_file(\%PARAMS);
if ( $rc > 0 ) { #got a good dataset
	# build image
	my $img = build_image( \%PARAMS );
	# print image
	print_image( $img, $PARAMS{"filename"} );
	# write web map html if requested
	if ( defined $PARAMS{"web"} and $PARAMS{"web"} ne "" ) {
		print print_imagemap( \%PARAMS );
	}
} else {
	die "Couldn't process the data in the input file!\n";
}

sub build_image {
	my $params = shift;
	my $img_x = $params->{"columns"} * $params->{"glyph_x"} +2;
							#+ (1 * $params->{"side_spacer"});
	
	print STDERR $params->{"total_entries"} / $params->{"columns"}, "\n";
	my $rows = int($params->{"total_entries"} / $params->{"columns"});
	$params->{"rows"} = $rows;

	my $img_y = 2 * $params->{"border"} +
							$params->{"total_entries"} / $params->{"columns"} * 
							$params->{"glyph_y"} + $params->{"glyph_y"};
	print STDERR "ROWS = $rows\n";

	$img_y = int($img_y);

	$params->{"img_x"} = $img_x;
	$params->{"img_y"} = $img_y;
	print STDERR "Image size will be $img_x by $img_y\n";

	my $img = new GD::Image( $img_x, $img_y );
	# allocate some colors
	my $lgray  = $img->colorAllocate(192,192,192);
	my $lgray2 = $img->colorAllocate(128,128,128);
	my $gray   = $img->colorAllocate(64,64,64);
	my $white  = $img->colorAllocate(255,255,255);
	my $yellow = $img->colorAllocate(255,255,0);
	my $black  = $img->colorAllocate(0,0,0);       
	my $pink   = $img->colorAllocate(255,0,255);      
	my $blue   = $img->colorAllocate(0,0,122);
	my $lblue  = $img->colorAllocate(0,128,255);
	my $green  = $img->colorAllocate(0,192,128);
	%COLORS = ( "white" => $white, "black" => $black, "lgray" => $lgray,
							"pink"  => $pink,  "gray" => $gray, "blue" => $blue,
							"lblue" => $lblue, "green" => $green, "yellow" => $yellow,
							"lgray2" => $lgray2);
	# make image interlaced (faster webloading)
	$img->interlaced('true');
	
	my $count = 1;
	my $colcount = 1;
	if ( defined $params->{"SHADE"} and $params->{"SHADE"} == 1 ) {
		draw_shaded_box( $img, $params );
	}

	foreach my $datum ( sort keys %{ $params->{"DATA_COORD"} } ) {
		if ( $count == 10 and $params->{"GRID"} ) {
			$count = 1;
			draw_kustu_glyph( $img,$datum,
				$params->{"DATA_COORD"}->{$datum},$params,1);
		} else {
			draw_kustu_glyph( $img,$datum,
				$params->{"DATA_COORD"}->{$datum},$params,0);
			$count++;
		}
		$colcount++;
		if ( $colcount > $params->{"columns"} ) {
			$colcount = 1;
		}
	}
	return $img;
}

sub draw_shaded_box {
	my( $img,$params ) = @_;
	my $shade = 0;
	my $x = 0;
	my $y = $params->{"border"};
	my $y_increment = $params->{"glyph_y"};
	for( my $r = 1; $r <= $params->{"rows"}+1; $r++ ) {
		if ( $shade ) {
			$img->filledRectangle( $x, $y, $params->{"img_x"},
				$y+$y_increment, $COLORS{ "lgray2"} );
		} else {
			$img->filledRectangle( $x, $y, $params->{"img_x"},
				$y+$y_increment, $COLORS{ "lgray"} );
		}
		if ( $shade ) { $shade = 0;} else { $shade = 1; }
		$y += $y_increment;
	}
}

sub swap_data {
	my($aref,$pos1,$pos2) = @_;
	if ( defined $pos1 and defined $pos2 ) {
		my $temp1 = $aref->[ $pos1 ];
		my $temp2 = $aref->[ $pos2 ];
		$aref->[ $pos1 ] = $temp2;
		$aref->[ $pos2 ] = $temp1;
	} else {
		my $temp1 = $aref->[ 0 ];
		my $temp2 = $aref->[ 1 ];
		$aref->[ 0 ] = $temp2;
		$aref->[ 1 ] = $temp1;
	}
	return $aref;
}

sub draw_kustu_glyph {
	my( $img,$datum,$coords,$params,$grid) = @_;
	my($x,$y) = (split(/,/, $coords))[0,1];
	
	my $ubox_x1 = $x+1;
	my $ubox_x2 = $ubox_x1 + $params->{"box_size"};
	my $ubox_y1 = $y + $params->{"top_spacer"};
	my $ubox_y2 = $ubox_y1 + $params->{"box_size"};
	my $lbox_x1 = $x+1;
	my $lbox_x2 = $lbox_x1 + $params->{"box_size"};
	my $lbox_y1 = $ubox_y2 + $params->{"bottom_spacer"};
	my $lbox_y2 = $lbox_y1 + $params->{"box_size"};

	my @data = split(/:/,$params->{"DATA"}->{$datum});
	if ( $params->{"SWAP"} ) {
		@data = @{ swap_data( \@data, 1, 2 ) };
	}

	#upper
	if ( $data[1] ne "" and $data[1] ne "A" and $data[2] eq "P" ) {
		# got a match
		if ( $params->{"match_rect"} ) {
			# show connected rectangle for the matching pair
			$img->filledRectangle( $ubox_x1, $ubox_y1, $lbox_x2, $lbox_y2,
				$COLORS{ $params->{"match_color"}} );
		} else {
			# user doesn't want a connected rect for matches, show indiv boxes
			$img->filledRectangle( $ubox_x1, $ubox_y1, $ubox_x2, $ubox_y2,
				$COLORS{ $params->{"ubox_border"}} );
			$img->filledRectangle( $ubox_x1+1, $ubox_y1+1, $ubox_x2 -1, $ubox_y2 -1,
				$COLORS{ $params->{"ubox_color"}} );
			$img->filledRectangle( $lbox_x1, $lbox_y1, $lbox_x2, $lbox_y2,
				$COLORS{ $params->{"lbox_border"}} );
			$img->filledRectangle( $lbox_x1+1, $lbox_y1+1, $lbox_x2 -1, $lbox_y2 -1,
				$COLORS{ $params->{"lbox_color"}} );
		}
	} elsif ( $data[1] ne "" and $data[1] ne "A" ) {
		# no secondary (protein) data, so just draw a top box
		$img->filledRectangle( $ubox_x1, $ubox_y1, $ubox_x2, $ubox_y2,
			$COLORS{ $params->{"ubox_border"}} );
		$img->filledRectangle( $ubox_x1+1, $ubox_y1+1, $ubox_x2 -1, $ubox_y2 -1,
			$COLORS{ $params->{"ubox_color"}} );
	} elsif ( $data[2] eq "P") {
		$img->filledRectangle( $lbox_x1, $lbox_y1, $lbox_x2, $lbox_y2,
			$COLORS{ $params->{"lbox_border"}} );
		$img->filledRectangle( $lbox_x1+1, $lbox_y1+1, $lbox_x2 -1, $lbox_y2 -1,
			$COLORS{ $params->{"lbox_color"}} );
	}
	else {
	}
	if ( $grid ) {
		$img->line( $ubox_x2+1, $y-2, $ubox_x2+1, $lbox_y2+2,
			$COLORS{ "gray"});
	}
}

sub build_ann_hash {
	my %annotation = ();
	open(IN, "<$gene_desc_file") or die "Couldn't open the gene descriptions file!";
	while(<IN>) {
		chomp;
		my @temp = split(/\t/, $_);
		$annotation{ $temp[0] } = $temp[1] . "!" . $temp[2];
	}
	close(IN);

	return \%annotation;
}

sub print_imagemap {
	my $params = shift;
	my $mapname = $params->{"web"};
	my $annotation = {};
	if ( $params->{"java"} ) {
		$annotation = build_ann_hash();
	}
	my $maphtml = "";
	my $line = "";
	my @buffer = ();
	$line =  qq!<map name="$mapname">!;
	push(@buffer, $line );
	$line = "";

	foreach my $datum ( sort keys %{ $params->{"DATA_COORD"} } ) {
		my @temp = split(/,/, $params->{"DATA_COORD"}->{ $datum });
		my $x1 = $temp[0];
		my $y1 = $temp[1];
		my $x2 = $x1 + $params->{"glyph_x"};
		my $y2 = $y1 + $params->{"glyph_y"};
		my $tag = qq!<area SHAPE=RECT COORDS="$x1,$y1,$x2,$y2" !;
		my $url = qq!HREF="http://coli.berkeley.edu/cgi-bin/ecoli/coli_entry.pl?seqid_out=$datum"!;
		if ( $params->{"java"} ) {
			my @temp = split(/!/, $annotation->{"$datum"});

			$url .= qq! onMouseOver="change('$datum','$temp[0]','$temp[1]')"!;
			$url .= qq! onMouseOut="clearform()"!;
		}
		$line = $tag . $url . qq! alt="$datum">!;
		push(@buffer, $line );
    $line = "";
	}
	$line = qq!</map>!;
	push(@buffer, $line );
	$line = "";
	$maphtml = join("\n",@buffer);
	return( $maphtml );
}

sub print_image {
	my $img = shift;
	my $filename = shift;
	$filename = $filename . ".png";
	open( OUT, ">$filename" ) or die "Problems opening $filename\n";
	print OUT $img->png;
	close(OUT);
}

sub process_file {
	my $params = shift;
	my $filename = $params->{"filename"};
	my $count = 0;
	open( IN, "<$filename" ) or die "Couldn't open file $filename\n";
	while(<IN>) {
		next if ( /^#/ );
		next if ( /^$/ );
		next if ( /^\s+$/ );
		chomp;
		my @temp = split( /\t/, $_ );
		shift(@temp);
		my $bnum = shift(@temp);
		$params->{"DATA"}->{ $bnum } = join(":", @temp);
		$count++;
	}
	$params->{"total_entries"} = $count;
	close(IN);
	print STDERR "Completed processing $filename: $count entries\n";

	# build data x,y coords
	my $xcursor = $params->{"border"}; 
	my $ycursor = $params->{"border"};
	my $col = 1;
	foreach my $datum ( sort keys %{ $params->{"DATA"} } ) {
		if ( $col > $params->{"columns"} ) {
			$col = 1; # reset col index
			$xcursor = $params->{"border"};
			$ycursor += $params->{"glyph_y"};
			$params->{"DATA_COORD"}->{ $datum } = "$xcursor,$ycursor";
			$xcursor += $params->{"glyph_x"};
			$col++;
		} else {
			$params->{"DATA_COORD"}->{ $datum } = "$xcursor,$ycursor";
			$xcursor += $params->{"glyph_x"};
			$col++;
		}
	}
	print STDERR "Completed calculating coordinates for image components\n";
	return $count;
}

sub parse_args {
	my $args = shift;
	my $params = shift;
	foreach my $arg ( @$args ) {
		if ( $arg =~ /=/ ) {
			my @temp = split(/=/, $arg );
			$params->{"$temp[0]"} = $temp[1];
		} elsif ( exists $params->{ "filename" } ) {
			print "Can only process one file!\n";
			print "filename is already set to ", $params->{ "filename" }, "\n";
			die "tried to set filename to $arg\n";
		} else {
			$params->{ "filename" } = "$arg";
		}
	}
	print STDERR "Completed processing parameters\n";
	# Check/Set defaults
	if ( not exists $params->{"columns"} ) {
		$params->{"columns"} = 100;
	}
	if ( not exists $params->{"box_size"} ) {
		$params->{"box_size"} = 6;
	}
	if ( not exists $params->{"match_color"} ) {
		$params->{"match_color"} = "green";
	}
	if ( not exists $params->{"match_rect"} ) {
		$params->{"match_rect"} = 0;
	}
	if ( not exists $params->{"ubox_color"} ) {
		$params->{"ubox_color"} = "blue";
	}
	if ( not exists $params->{"lbox_color"} ) {
		$params->{"lbox_color"} = "yellow";
	}
	if ( not exists $params->{"ubox_border"} ) {
		$params->{"ubox_border"} = "blue";
	}
	if ( not exists $params->{"lbox_border"} ) {
		$params->{"lbox_border"} = "yellow";
	}
	if ( not exists $params->{"side_spacer"} ) {
		$params->{"side_spacer"} = 1;
	}
	if ( not exists $params->{"top_spacer"} ) {
		$params->{"top_spacer"} = 1;
	}
	if ( not exists $params->{"bottom_spacer"} ) {
		$params->{"bottom_spacer"} = 1;
	}
	if ( not exists $params->{"border"} ) {
		$params->{"border"} = 2;
	}
	if ( not exists $params->{"GRID"} ) {
		$params->{"GRID"} = 0;
	}
	if ( not exists $params->{"SHADE"} ) {
		$params->{"SHADE"} = 1;
	}
	if ( not exists $params->{"SWAP"} ) {
		$params->{"SWAP"} = 0;
	}
	# calculate glyph sizes
	$params->{"glyph_x"} = $params->{"box_size"} + $params->{"side_spacer"} + 1;
	$params->{"glyph_y"} = 2 * $params->{"box_size"} +
			2 * $params->{"bottom_spacer"} + 2 * $params->{"top_spacer"};
	print STDERR "Completed checking/setting defaults\n";

	return;
}

sub usage {
print <<EOT;
generate_boximage.pl filename

EOT
die "Incorrect/Missing Parameters\n";
}

sub print_java {
	print <<EOT;

<script language="JavaScript" type="text/javascript">
function clearform() {
	document.geneinfo.bnum.value = ''	
	document.geneinfo.gene.value = ''
	document.geneinfo.func.value = ''
}
function change(bnum, gene, func ) {
	document.geneinfo.bnum.value = bnum	
	document.geneinfo.gene.value = gene
	document.geneinfo.func.value = func
}
</script>

EOT
}
